#!/bin/sh
#
# Cross Control - Device updater
#
# Update existing XS/CrossCore/CrossLink devices
# 
# Below V is updater identification, increment with update (might become usefull someday)
V=A8.3
DEVICE="Cross XA1"
# 110930 TEK, A8.3:Cross XA1 - mtd8 partition fix
# 100518 TTH, A8.2:Cross XA1 - CC+under-temp-fix 
# 100322 TTH, A8.1:Cross XA1 - Bootstrap flash oops
# 100318 TTH, A8:Cross XA1 - With FPGA
# 1003xx TTH, A7:Cross XA1 - Cleaned and structured, prepared for FPGA
# 100224 TTH, A6:Cross XA1 - RW-mode protection
# 100127 TTH, A5:Cross XA1 - Using backup-system term
# 091120 TTH, A4:Cross XA1 - reboot-tune-up
# 091117 TTH, A3:Cross XA1 - With bootloader
# 091106 TTH, A2:Cross XA1 - Secondary System support
# 091026 TTH, A1:Cross XA1 - Proto environment version
#    -- branch point for new device --
# 090422 MNK, Now with ramimage system support as well
# 090211 TTH, Now with system image and separate files and CF image
# 090202 TTH, Now with bootloader
# 0812.. TTH, 
#

# Help usage info
show_help()
{
if [ ! "x$1" == "x" ] ; then 
    echo -e "\nErr:"$1"\n" ; 
    echo -e " Run with -h for help and details.\n"
    exit 4
fi
cat <<EOF
 Reflash $DEVICE device with new firmware, distribution and addons.

Usage: `basename $0` <-p|-s> [-b file] [-k file] [-r file] [[-c file] [-d]]
Usage: `basename $0` -l mcs-file

  Choice of update target:
   -p   Update Primary side images (default)
   -s   Update Backup-System images
   -l   Update FPGA Logic code (not together with others)
   -c   Addons package to extract to CF (add -p or -s if want both) 

  Default names used unless overwritten with command line:
   -b   Bootloader:  cross-xa1_bootstrap.bin
   -k   Kernel:      cross-xa1_kernel.bin
   -r   Rootfs:      cross-xa1_rootfs.bin

   -d   When extracting CF-pack, clear included folders
   -f   Force-flag disables all user interaction, use for scripting.
   -m   Disable MD5 sum matching - DANGEROUS
   -u   Clear /usr/local/*
   -2   Do reboot-rescue instead of reboot after flashing 

   -v   Verbose messages
   -q   Quiet messages

 NOTE: Shutdown all excess applications BEFORE executing this update!
       Stop all file-system access and prevent preliminary power loss.

EOF
 
 exit 4
}

## Conditional Leveled echos, level 1=always, 2=normally, 3=in verbose
# args: <level> <what to echo>
decho (){
 EL=$1
 shift
 if [ $EL -lt $DL ] ; then
       ./busybox echo "$*"
 fi
}
dechon (){
 EL=$1
 shift
 if [ $EL -lt $DL ] ; then
     ./busybox echo -n "$*"
 fi
}
dechoe (){
 EL=$1
 shift
 if [ $EL -lt $DL ] ; then
     ./busybox echo -e "$*"
 fi
}
dechoen (){
 EL=$1
 shift
 if [ $EL -lt $DL ] ; then
     ./busybox echo -en "$*"
 fi
}


# Prepare things for FPGA update
prepare_fpga()
{
  ferr=0
  rm -rf /tmp/dev/fpgaflash
  rmmod fpgaspi 2>/dev/null
  (insmod /tmp/fpgaspi.ko block=0xCC verbose=0 2>/dev/null || modprobe fpgaspi block=0xCC) 2>&1 >/dev/null && \
  mknod /tmp/dev/fpgaflash c `grep cross_fpgaflash /proc/devices | cut -d\  -f1` 0 || \
  ferr=1

  if [ $ferr -eq 1 ] ; then
    decho 1 "ERROR: Something wrong while preparing for FPGA update. - Quiting.."
    exit 3
  fi

}

# Do the FPGA
flash_fpga()
{
    echo "Erasing and Flashing FPGA" && \
    ./mcstool -f $TMPFBIN && \
    echo "Verifying the flashed data" && \
    ./mcstool -r $TMPFBIN.out `ls -l $TMPFBIN | tr -s \  | cut -d\  -f5` >/dev/null && \
    ./mcstool -m $TMPFBIN.out readback.mcs >/dev/null && \
    dos2unix readback.mcs && \
    dos2unix $FIMG && \
    diff $FIMG readback.mcs -qs 
    if [ $? -ne 0 ] ; then
        echo -e   "Error: Something went wrong, during the flash !"
        echo -e "\n YOUR FPGA FLASH IS NOW CORRUPT!  FIX IT BEFORE CUTTING POWER!\n"
        echo -e   " If you power off, the device might not reboot. !\n "
        exit 1
    fi
    echo -e "\nResult: FPGA code flashed and verified successfully."
    echo "        The new code is used after next full power off! (not reboot)"
    exit 0  ## Exit with success here, no need to reboot
}



checkFile()
{
 FN=$1
 if [ ! -f $MD5 ] ; then
  if [ $md5 -eq 0 ] ; then 
   decho 3 " WARNING: MD5 file: Not found but Ignoring with \"-m\"" 
   return 0
  else 
   decho 1 "   ERROR: MD5 file: $MD5: Not found. - Quiting.."
   exit 5
  fi
 fi
 dechon 2  "      Checking MD5: $FN: "
 
 (grep -E " $FN$" $MD5 || echo "" ) | md5sum -c - -s 

 if [ $? -ne 0 ] ; then 
  # error
  if [ $md5 -eq 0 ] ; then 
   decho 2 "FAILED! - Ignoring with \"-m\"" 
  else 
   decho 2 "FAILED! - Quiting.." 
   exit 5;
  fi 
 else
  decho 2 "OK"
 fi

}

check_environment() {

 # Show only help if run outside of XA1
 grep -q "Atmel.*AT91SAM" /proc/cpuinfo
 if [ $? -ne 0 ] ; then show_help "This works only when used at $DEVICE device"; fi

 if [ `id -u` -ne 0 ] ; then show_help "This works ONLY when used as root -user!" ; fi

 # Normal side or Secondary system?
 ROOTUBIDEV=`grep :rootfs /proc/mounts | grep -o ubi.`

 this_side=0 #unknown
 if [ "$ROOTUBIDEV" == "ubi1" ] ; then this_side=1 ; # normal
 elif [ "$ROOTUBIDEV" == "ubi0" ] ; then this_side=2 ; # second
 else
   echo "Unknown run environment... - Quiting.."
   exit 9
 fi
 # Where are we
 PWD=`pwd`
 if [ ! ${PWD:0:4} == "/tmp" ] ; then echo "WARNING: Executing outside of /tmp, expecting problems..." ; fi
 # Read only local
 touch temporary_file_of_no_purpose 2>/dev/null
 if [ $? -ne 0 ] ; then show_help "Unable to write to local folder. Quiting.." ; 
 else
   rm -f temporary_file_of_no_purpose
 fi
}

set_defaults() {

# Default image names
MD5=cross-xa1.md5sum
BIMG=cross-xa1_bootstrap.bin
KIMG=
RIMG=
FIMG=
KIMG1=cross-xa1_kernel.bin
RIMG1=cross-xa1_rootfs.bin
KIMG2=cross-xa1-bs_kernel.bin
RIMG2=cross-xa1-bs_rootfs.bin
CIMG=
SIMG=

TMPFBIN=/tmp/fpga.bin

# Device names
BDEV=
KDEV=
RDEV=
FDEV=fpga
KDEV1=mtd0
RDEV1=mtd3

KDEV2=mtd1
RDEV2=mtd2

bi=1
ki=1
ri=1
fi=0

si=0
ni=0

ci=0
de=0

usrclear=0

ask=1
md5=1
update_side=1  # default to main system
update_this=1  # allow it quietly always (?)

DL=3 # Debug/verbose level

}

# Return 1 if argument exists
# Return 0 if argument not exist
get_argument() {
 A=$1
 RV=0
 shift
 CB=`getopt -uq -- $A "$@"`
 CB=`echo $CB| sed "s/--.*//"`
 if [ ! -z "$CB" ] ; then RV=1 ; fi
}

# Return value of argument
# Return 0 if no/empty
get_argvalue() {
 A=$1
 shift
 #echo ";$@;"
 CB=`getopt -uq -- $A: "$@" | sed "s/^ *[^ ]* \(.*\) -.*/\1/" | sed "s/\-.*//"`
 if [ -z $CB ] ; then  
   RV=0
 fi
 RV=$CB
}

# Return filename argument value and check for existence.
# Quit on not available
get_filearg() {
 A=$1
 shift
 #echo ";$@;"
 CB=`getopt -uq -- $A: "$@" | sed "s/^ *[^ ]* \(.*\) -.*/\1/" | sed "s/\-\-.*//"`
 if [ -z $CB ] ; then  
   show_help "Option -$A requires a file name argument."
 fi
 if [ ! -f $CB ] ; then
   show_help "File for -$A ($CB) not found!"
 fi
 RV=$CB
}

# Return filename argument valu and check for existence.
# If argument not avail, ignore.
get_optfilearg() {
 get_argument $@
 if [ $RV -eq 1 ] ; then get_filearg $@ ; else RV="" ; fi
}

# Read all arguments and set variables accordingly
read_arguments() { 

 if [ $# -gt 0 ]
 then

  get_argument q $@
  if [ $RV -eq 1 ] ; then DL=2; fi
  get_argument v $@
  if [ $RV -eq 1 ] ; then DL=4; fi

  # H elp
  get_argument h $@
  if [ $RV -eq 1 ] ; then show_help; fi

  # F orce
  get_argument f $@
  if [ $RV -eq 1 ] ; then ask=0; fi

  # M d5 check
  get_argument m $@
  if [ $RV -eq 1 ] ; then md5=0; fi

  get_argument d $@
  if [ $RV -eq 1 ] ; then de=1; fi

  # U sr/local delete check
  get_argument u $@
  if [ $RV -eq 1 ] ; then 
    usrclear=1; 
    de=0 # no more need for this
  fi

  # Make next reboot -> reboot-rescue
  get_argument 2 $@
  if [ $RV -eq 1 ] ; then 
    echo "1 0" >/dev/factory_reset
  fi

  # Make next reboot-rescue -> reboot
  get_argument 1 $@
  if [ $RV -eq 1 ] ; then 
    echo "0 0" >/dev/factory_reset
  fi

  get_optfilearg c $@
  if [ ! -z "$RV" ] ; then 
    ci=1
    bi=0
    ki=0
    ri=0
    fi=0
    if [ ${RV:0:1} == "/" ] ; then
      CIMG="$RV"
    else
      CIMG="/tmp/$RV"
    fi

    update_side=0;  # remove default 
  fi

  get_argument s $@
  if [ $RV -eq 1 ] ; then 
    update_side=2; 
    bi=1
    ki=1
    ri=1
    fi=0
  fi

  get_argument p $@
  if [ $RV -eq 1 ] ; then 
    update_side=1;
    bi=1
    ki=1
    ri=1
    fi=0
  fi

  get_argument t $@
  if [ $RV -eq 1 ] ; then update_this=1; fi

  # Check for overrided filenames
  get_optfilearg b $@
  if [ ! -z "$RV" ] ; then BIMG="$RV" ; fi
  get_optfilearg k $@
  if [ ! -z "$RV" ] ; then KIMG="$RV" ; fi
  get_optfilearg r $@
  if [ ! -z "$RV" ] ; then RIMG="$RV" ; fi

  get_optfilearg l $@
  if [ ! -z "$RV" ] ; then
    FIMG="$RV" 
    update_side=4
    bi=0
    ki=0
    ri=0
    fi=1
  fi




fi # params exist


} #readarguments

# Display info if..
show_bootstrap() {
 if [ $bi -eq 1 ] ; then
  if [ -f $BIMG ] ; then
   decho 1 "  Requested action: Bootstrap update"
   decho 3 "   Bootstrap-image: $BIMG"
  else
   # no Bootstrap file, think its ok.
   bi=0
  fi
 fi

}

# Display preliminary info of what we are about to do
show_request() {

# sides ok?
 if [ $this_side -eq 1 ] ; then
  decho 1 " Now running on   : normal side"
 else
  decho 1 " Now running on   : BACKUP system"
 fi

 if [ $ci -eq 1 ] ; then
  decho 1  "  Requested action: Addon install"
  decho 3  "   Used file image: $CIMG"
 fi
 
 if [ $update_side -eq 4 ] ; then
  decho 1  "  Requested action: FPGA update"
  decho 2  "   Used file image: $FIMG"
 elif [ $update_side -eq 2 ] ; then
  if [ -z $KIMG ] ; then KIMG=$KIMG2; fi
  if [ -z $RIMG ] ; then RIMG=$RIMG2; fi
  KDEV=$KDEV2  
  RDEV=$RDEV2
  show_bootstrap
  dechoe 1 "  Requested action: BACKUP system update"
  dechoe 3 "      kernel-image: $KIMG"
  dechoe 3 "        root-image: $RIMG"

  if [ $this_side -ne 1 ] ; then
   if [ $update_this -eq 1 ] ; then
     decho 3 "Active side update: BY FORCE"
   else
     echo "Active side update: Denyed.";
     exit 7
   fi
  fi


 elif [ $update_side -eq 1 ] ; then
  if [ -z $KIMG ] ; then KIMG=$KIMG1; fi
  if [ -z $RIMG ] ; then RIMG=$RIMG1; fi
  KDEV=$KDEV1
  RDEV=$RDEV1
  show_bootstrap
  dechoe 1 "  Requested action: NORMAL system update"
  dechoe 3 "      kernel-image: $KIMG"
  dechoe 3 "        root-image: $RIMG"

  if [ $this_side -ne 2 ] ; then
   if [ $update_this -eq 1 ] ; then
     decho 3 "Active side update: BY FORCE"
   else
     echo "Active side update: Denyed."
     exit 7
   fi
  fi
   
#  else
#  show_help "Update side not selected. Exit."
#  exit 8
fi 

}

# Acquire sizes of partitions.
get_sizes() {
if [ ! -z "$KDEV" ] ; then
SIZEPK=`grep $KDEV /proc/mtd | sed -r "s/.* ([0-9a-f]*) ([0-9a-f]*) \"(.*)\"/0x\1/"`
fi
if [ ! -z "$RDEV" ] ; then
SIZEPR=`grep $RDEV /proc/mtd | sed -r "s/.* ([0-9a-f]*) ([0-9a-f]*) \"(.*)\"/0x\1/"`
fi
}

check_addon_image() {
 if [ ! -f $CIMG ] ; then
  echo "ERROR: Bad image? : $CIMG: Not found, skipping.."
  ci=0
 else 
  checkFile $CIMG
  dechon 2 "    Checking addon: $CIMG: "
  tar -tf $CIMG > /tmp/cflist
  if [ $? -ne 0 ] ; then
    decho 1 "Addon is corrupt. Quiting.."
    exit 2
  else 
    DIRS=`cat /tmp/cflist | grep "^[^\/]*\/$" | tr -d \/ `
    decho 2  "OK"
    decho 3  "    addon includes: " `echo $DIRS ` 
  fi
 fi 
}

# CF Addons package install  // 100304, not verified.
addon_install() {

 ./busybox echo "Extracting addon package.."
 if [ $de -eq 1 ] ; then
   ./busybox echo ".. after clearing included folders."
 fi

 cd /usr/local || exit 9

 if [ $de -eq 1 ] ; then
   rm -rf $DIRS
 fi

 tar -xf $CIMG

 echo -e "\n Addons image installed\n"
 cd /tmp || exit 8

}

# Verify file size range
# args: <Filename> <minsize> <maxsize>
check_file_size() {
 FN=$1
 MIN=$2
 MAX=$3
 SIZEI=`ls -l $FN | tr -s \  | cut -d\  -f5`
 if [ $((SIZEI)) -lt $((MIN)) ]
  then
   echo " ERROR: Image file: $FN: Smaller than expected. - Quiting.."
   exit 1
  fi
  if [ $((SIZEI)) -gt $((MAX)) ]
  then
   echo " ERROR: Image file: $FN: Larger than expected. - Quiting.."
   exit 1
  fi
}



# Verify image correctness
check_bootstrap_image() {
 if [ ! -f $BIMG ] ; then
  echo " ERROR: Image file: $BIMG: Not found. Quiting.."
  bi=0
  exit 1
 else 
  checkFile $BIMG
  check_file_size $BIMG 5000 15000
 fi 

}

# Verify image correctness
check_fpga_image() {
 if [ ! -f $FIMG ] ; then
  echo " ERROR: Image file: $FIMG: Not found. Quiting.."
  fi=0
  exit 1
 else 
  checkFile $FIMG
  ./mcstool $FIMG $TMPFBIN >/dev/null
  if [ $? -ne 0 ] ; then 
   echo " ERROR: Image file: $FIMG: Format error. Quiting.."
   fi=0
   exit 1
  fi
 fi 

}

# Verify image correctness
check_kernel_image() {
 if [ ! -f $KIMG ] ; then
   ki=0
   if [ $bi -eq 1 ] ; then
     # allow bootstrap only update
     echo " WARNING:    Image: $KIMG: Not found. - Skipping.."
   else
     echo " ERROR: Image file: $KIMG: Not found. - Quiting.."
     exit 1
   fi
 else 
  checkFile $KIMG
  check_file_size $KIMG 120000 $SIZEPK
 fi 
}

# Verify image correctness
check_root_image() {
 if [ ! -f $RIMG ] ; then
   ri=0
   if [ $bi -eq 1 ] ; then
     # allow bootsrap only update
     echo " WARNING:    Image: $RIMG: Not found. - Skipping.."
   else
     echo " ERROR: Image file: $RIMG: Not found. - Quiting.."
     exit 1
   fi
 else 
  checkFile $RIMG
  # known sizes 091106 bs-root: 9699328, main-root: 21102592
  check_file_size $RIMG 9000000 $SIZEPR
 fi 
}
# Verify files, iff
check_files() {
 get_sizes
 if [ $ci -eq 1 ] ; then
   check_addon_image
 fi

  if [ $fi -eq 1 ] ; then
    check_fpga_image
  fi
  if [ $bi -eq 1 ] ; then
    check_bootstrap_image
  fi
  if [ $ki -eq 1 ] ; then
    check_kernel_image
  fi
  if [ $ri -eq 1 ] ; then
    check_root_image
  fi

 
}


# Args: <filename> <source> <test-cmd>
copyhere ()
{
 if [ ! -x $1 ] ; then
  cp -f $2 .
 fi
 if [ ! -x $1 ] ; then
  echo "Error: Local version of '$1' missing. - Quiting.."
  exit 1
 fi
 ./$3 >/dev/null
 if [ $? -ne 0 ] ; then
  echo "Error: Nonworking version of local '$1'. - Quiting.."
  exit 2
 fi
}

# Init-like
prepare_environment() {
prepare_services

rm -f /tmp/mtd*
mknod /tmp/mtd0 c 90 0 && 
mknod /tmp/mtd1 c 90 2 && 
mknod /tmp/mtd2 c 90 4 &&
mknod /tmp/mtd3 c 90 6 && 
mknod /tmp/mtd4 c 90 8 && 
mknod /tmp/mtd6 c 90 12 &&
mknod /tmp/mtd7 c 90 14 &&
mknod /tmp/mtd8 c 90 16 || show_help "Prepare failed"
if [ $bi -eq 1 ] ; then search_bootstrap ; fi
if [ $fi -eq 1 ] ; then prepare_fpga ; fi
sleep 1
}


# Check which of the potential devs is bootstrap 
search_bootstrap(){
if [ $bi -eq 1 ] ; then 
    echo "Setting fix bootstrap"
    BDEV=mtd8

    if [ -z $BDEV ] ; then
	echo "Unable to locate correct bootstrap device, skipping bootstrap update."
	bi=0
    fi

fi
}

# Check for required tools
prepare_tools(){
 copyhere busybox /bin/busybox "busybox echo 'good'"
 copyhere flash_eraseall /usr/sbin/flash_eraseall "flash_eraseall --version"
 copyhere nandwrite /usr/sbin/nandwrite "nandwrite --version"
 if [ $fi -eq 1 ] ; then copyhere mcstool "/usr/sbin/mcstool" "mcstool" ; fi
}

# Check constrains for actions
check_limits(){
if [ `expr $ki + $ri` -eq 1 ]
then
 echo -e "WARNING: Kernel and Rootfs, only when both available - Skipping.."
 ki=0
 ri=0
fi
# 
if [ `expr $bi + $ki + $ri + $ci + $fi` -eq 0 ]
then
 echo "ERROR: Cant flash without accepted set of images. - Quiting.."
 exit 1
fi
}

# Listing parts of actions to do 
show_effects() {
echo
if [ $bi -eq 1 ]
then
 decho 1 " Updating: bootstrap  "
fi
if [ $fi -eq 1 ]
then
 decho 1 " Updating: FPGA   "
 decho 2 "           Dangerous if failed"
fi
if [ $usrclear -eq 1 ]
then
 decho 1 " Clearing: /usr/local/  ALL FILES WILL BE LOST."
fi
if [ $ci -eq 1 ]
then
 if [ $de -eq 1 ] ; then
   decho 1 "  Install: $CIMG  after clearing existing folders."
 else
   decho 1 "  Install: $CIMG  Overrides existing files."
 fi

fi
if [ $ki -eq 1 ]
then
 decho 1 " Updating: kernel    "
fi
if [ $ri -eq 1 ]
then
 decho 1 " Updating: root-fs      ALL FILES ON ROOT-FS WILL BE LOST"
 if [ $usrclear -eq 1 ] ; then
   decho 2 "           Files on /media/cf unaffected" 
 else
   decho 2 "           Files on /usr/local/ and /media/cf unaffected" 
 fi
fi

}

prepare_services(){
 decho 3 "         Preparing: just a minute.."
 if [ -e /etc/init.d/logrotate ] ; then
 /etc/init.d/logrotate stop 2>&1 >/dev/null; fi
 # protect from scarey RW mode
 mount none / -oremount,ro
 #/etc/rcS.d/S90powerd stop
 sleep 1
}

# Final Confirmation
confirm_user(){
 if [ $ask -eq 1 ] ; then
  echo -e "\n User verification: ARE YOU SURE ?\n"
  echo -e   "   - If yes, press Enter to continue."
  echo -e   "   - IF NOT, press CTRL+C now to cancel update."
  decho 2   "  This is you last change to do cancel!"
  dechoe 1  "\n** IF YOU CONTINUE, DO NOT INTERRUPT THE PROCESS UNTIL READY **" 
  read a
 else   
  echo -e "\n User verification: Skipped, auto continue.."
 fi
 dechoe 2 "\n         Now doing: Requested reprogramming:"

 dechoe 1 "\n * DO NOT BREAK PROCESS OR SHUTDOWN THE UNIT before update is complete. *\n"
 sleep 1

}


# Usr local clear
clear_usrlocal() {
 ./busybox echo "Cleaning /usr/local .."
  rm -rf /usr/local/*
}


##
## FLASH EM
##
flash_bootstrap() {
./busybox echo "Copying new bootstrap image.."
./busybox dd if=$BIMG of=/tmp/$BDEV conv=sync
}

flash_kernel() {
./busybox echo "Erasing old kernel.."
./flash_eraseall /tmp/$KDEV
./busybox echo "Copying new kernel image.."
./nandwrite -p /tmp/$KDEV $KIMG
}

flash_root() {
./busybox echo "Erasing old root.."
./flash_eraseall /tmp/$RDEV
./busybox echo "Copying new root-fs image.."
./nandwrite /tmp/$RDEV $RIMG
}



do_reboot(){

# All done, rebooooot
./busybox echo "Completed. Rebooting.."
./busybox reboot

}

# YEY...

echo "===== CrossControl: Device Update ===== "$V

set_defaults

read_arguments $@

check_environment 

prepare_tools

show_request

check_files

check_limits

prepare_environment

show_effects

confirm_user

if [ $usrclear -eq 1 ] ; then
 clear_usrlocal
fi

if [ $ci -eq 1 ] ; then
 addon_install
fi

if [ $fi -eq 1 ] ; then
 flash_fpga
 exit 1
fi

if [ $bi -eq 1 ] ; then
 flash_bootstrap
fi

if [ $ki -eq 1 ] ; then
 flash_kernel
fi

if [ $ri -eq 1 ] ; then
 flash_root
fi

do_reboot

./busybox echo -e "\nEnd of all actions, H o l d    o n    t i g h t \n"
